#pragma once

#include "System.h"

// #include "../../StringBuffer.h"

namespace Pomi {

class I2C {
    
    public:
    
        // static inline StringBuffer<80> sda_str;
        // static inline StringBuffer<80> scl_str;
        
        I2C() = delete;
        I2C(const I2C&) = delete;
        I2C& operator =(const I2C&) = delete;
        
        static inline unsigned char read() {
            return 0;
        }
        
        static inline void write(unsigned char address, unsigned char data) {
            // printf("[0x%.2x] <= 0x%.2x\n", address, data);
            
            // sda_str.replace(0, sda_str.length(), "");
            // scl_str.replace(0, scl_str.length(), "");
            
            _start();
            
            _sendFrame(address & ~1);
            _sendFrame(data);
            
            _stop();
            
            // printf("data: %s\n", static_cast<const char*>(sda_str));
            // printf("clock:%s\n", static_cast<const char*>(scl_str));
        }
    
    private:
    
        static constexpr unsigned int SDA = 4;
        static constexpr unsigned int SCL = 5;
        
        static inline void _waitMicros(unsigned int delay) {
            unsigned int start = System::currentTimeMicros();
            while(System::currentTimeMicros() - start < delay);
        }
        
        static inline void _start() {
            unsigned int* DIR = reinterpret_cast<unsigned int*>(0xa0002000);
            volatile unsigned char* P0 = reinterpret_cast<unsigned char*>(0xa0000000);
            
            P0[SDA] = 1;
            P0[SCL] = 1;
            
            // Set SDA and SCL as outputs
            DIR[0] |= (1<<SDA) | (1<<SCL);
            
            // sda_str.append("^"); scl_str.append("^");
            
            _waitMicros(1);
            // sda_str.append("\\"); scl_str.append("^");
            P0[SDA] = 0;
            _waitMicros(1);
            // sda_str.append("_"); scl_str.append("\\");
            P0[SCL] = 0;
        }
        
        static inline void _stop() {
            unsigned int* DIR = reinterpret_cast<unsigned int*>(0xa0002000);
            volatile unsigned char* P0 = reinterpret_cast<unsigned char*>(0xa0000000);
            
            // sda_str.append("_"); scl_str.append("_");
            P0[SDA] = 0;
            // sda_str.append("_"); scl_str.append("/");
            P0[SCL] = 1;
            _waitMicros(1);
            // sda_str.append("/"); scl_str.append("^");
            P0[SDA] = 1;
            
            // Set SDA and SCL as inputs
            DIR[0] &= ~(1<<SDA) & ~(1<<SCL);
        }
        
        static inline bool _sendFrame(unsigned char value) {
            unsigned int* DIR = reinterpret_cast<unsigned int*>(0xa0002000);
            volatile unsigned char* P0 = reinterpret_cast<unsigned char*>(0xa0000000);
            
            // std::uint8_t data = 0;
            
            // Send out 8 bits, most significant bit first
            for(std::uint32_t bit = 8; bit > 0; --bit) {
                _waitMicros(1);
                // sda_str.append(data ? (((value >> (bit - 1)) & 1) ? "^" : "\\") : (((value >> (bit - 1)) & 1) ? "/" : "_")); scl_str.append("_");
                // data = (value >> (bit - 1)) & 1;
                P0[SDA] = (value >> (bit - 1)) & 1;
                // sda_str.append(data ? "^" : "_"); scl_str.append("/");
                P0[SCL] = 1;
                _waitMicros(1);
                // sda_str.append(data ? "^" : "_"); scl_str.append("\\");
                P0[SCL] = 0;
            }
            
            // Set SDA as input and ask for the ACK bit
            DIR[0] &= ~(1<<SDA);
            // sda_str.append("?"); scl_str.append("/");
            P0[SCL] = 1;
            _waitMicros(1);
            
            // Read ACK bit and set SDA as output
            bool ack = !P0[SDA];
            // sda_str.append("?"); scl_str.append("\\");
            P0[SCL] = 0;
            DIR[0] |= (1<<SDA);
            
            return ack;
        }
};

} // namespace Pomi
