#pragma once

#include <SDL2/SDL.h>
#include <PokittoSimulator.h>

namespace Pomi {
namespace Sim {

class Buttons {
    
    public:
    
        Buttons() = delete;
        Buttons(const Buttons&) = delete;
        Buttons& operator =(const Buttons&) = delete;
        
        static inline void poll() {
            SDL_Event event;
            SDL_PollEvent(&event);
            if(event.type == SDL_KEYDOWN) {
                switch(event.key.keysym.sym) {
                    case SDLK_LEFT: _states |= 1<<0; break;

                    case SDLK_RIGHT: _states |= 1<<1; break;

                    case SDLK_UP: _states |= 1<<2; break;

                    case SDLK_DOWN: _states |= 1<<3; break;

                    case SDLK_a: _states |= 1<<4; break;

                    case SDLK_b: case SDLK_s: _states |= 1<<5; break;

                    case SDLK_c: case SDLK_d: _states |= 1<<6; break;

                    case SDLK_ESCAPE:
                        Pokitto::Simulator::killSound();
                        Pokitto::Simulator::CleanUp();
                        exit(0);
                }
            }
            else if(event.type == SDL_KEYUP) {
                switch(event.key.keysym.sym) {
                    case SDLK_LEFT: _states &= ~(1<<0); break;

                    case SDLK_RIGHT: _states &= ~(1<<1); break;

                    case SDLK_UP: _states &= ~(1<<2); break;

                    case SDLK_DOWN: _states &= ~(1<<3); break;

                    case SDLK_a: _states &= ~(1<<4); break;

                    case SDLK_b: case SDLK_s: _states &= ~(1<<5); break;

                    case SDLK_c: case SDLK_d: _states &= ~(1<<6); break;
                }
            }
            else if(event.type == SDL_QUIT) {
                Pokitto::Simulator::killSound();
                Pokitto::Simulator::CleanUp();
                exit(0);
            }
        }
        
        static inline bool a() { return _states & (1<<4); }
        static inline bool b() { return _states & (1<<5); }
        static inline bool c() { return _states & (1<<6); }
        static inline bool up() { return _states & (1<<2); }
        static inline bool down() { return _states & (1<<3); }
        static inline bool left() { return _states & (1<<0); }
        static inline bool right() { return _states & (1<<1); }
    
    private:
    
        static inline std::uint8_t _states = 0;
    
};

} // namespace Sim
} // namespace Pomi
