#pragma once

#include <cstdint>
#include <utility>
#include <PokittoSimulator.h>

namespace Pomi {
namespace Sim {

class Display {
    
    public:
    
        Display() = delete;
        Display(const Display&) = delete;
        Display& operator =(const Display&) = delete;
        
        enum class Direction { HORIZONTAL = 0x1038, VERTICAL = 0x1030 };
        
        static inline void setDirection(Direction dir) {
            switch(dir) {
                case Direction::HORIZONTAL:
                    if(_row != &_x) {
                        _row = &_x;
                        _col = &_y;
                        std::swap(_row_first, _col_first);
                        std::swap(_row_last, _col_last);
                    }
                    break;

                case Direction::VERTICAL:
                    if(_row != &_y) {
                        _row = &_y;
                        _col = &_x;
                        std::swap(_row_first, _col_first);
                        std::swap(_row_last, _col_last);
                    }
                    break;
            }
        }
        
        static inline void setX(unsigned int x) {
            _x = x;
        }
        
        static inline void setY(unsigned int y) {
            _y = y;
        }
        
        static inline void beginStream() {}
        
        static inline void writeData(unsigned int data) {
            _data = data;
            writeData();
        }
        
        static inline void writeData() {
            if((_x < SIMW) && (_y < SIMH)) {
                Pokitto::Simulator::gfxbuf[_x + (_y*SIMW)] = _data;
            }

            *_row += 1;
            if(*_row > _row_last) {
                *_row = _row_first;
                *_col += 1;
                if(*_col > _col_last) {
                    *_col = _col_first;
                }
            }
        }
        
        static inline void enterReadMode(unsigned int x, unsigned int y) {
            _x = x;
            _y = y;
        }
        
        static inline void leaveReadMode() {}
        
        static inline unsigned int readBGR565() {
            if((_x < 0) || (_x >= SIMW) || (_y < 0) || (_y >= SIMH)) return 0;

            unsigned int _data = Pokitto::Simulator::gfxbuf[_x + (_y*SIMW)];
            return ((_data & 0x1f) << 11) | (_data & 0x7e) | (_data >> 11);
        }
        
        static inline void setXRange(unsigned int first, unsigned int last) {
            if(_row == &_x) {
                _row_last = last <= SIMW-1 ? last : SIMW-1;
                _row_first = first <= _row_last ? first : _row_last;
            }
            else {
                _col_last = last <= SIMW-1 ? last : SIMW-1;
                _col_first = first <= _col_last ? first : _col_last;
            }
        }
        
        static inline void setYRange(unsigned int first, unsigned int last) {
            if(_col == &_y) {
                _col_last = last <= SIMH-1 ? last : SIMH-1;
                _col_first = first <= _col_last ? first : _col_last;
            }
            else {
                _row_last = last <= SIMH-1 ? last : SIMH-1;
                _row_first = first <= _row_last ? first : _row_last;
            }
        }

    private:
    
        static inline unsigned int _x = 0;
        static inline unsigned int _y = 0;
        static inline unsigned int* _row = &_x;
        static inline unsigned int* _col = &_y;
        static inline unsigned int _data = 0;

        static inline unsigned int _x_first = 0;
        static inline unsigned int _x_last = SIMW-1;
        static inline unsigned int _y_first = 0;
        static inline unsigned int _y_last = SIMH-1;

        static inline unsigned int _row_first = _x_first;
        static inline unsigned int _row_last = _x_last;
        static inline unsigned int _col_first = _y_first;
        static inline unsigned int _col_last = _y_last;
};

} // namespace Sim
} // namespace Pomi
